from . import analysis
from . import conversion
from . import fractal
from . import genetic
from . import harmony
from . import loop
from . import minimalism
from . import player
from . import rhythm
from . import score
from . import utils
from . import walk

#from importlib.util import find_spec
#import warnings
#tensorflow_spec = find_spec('tensorflow')
#if tensorflow_spec is not None:
#    from . import djai
#else:
#    warnings.warn("Tensorflow is not installed. Djai cannot be imported.")

__version__ =  "0.2.1"